import copy
import warnings
from collections import UserList, UserDict

import numpy as np
import scipy.linalg


class BasePulse:
    """
    Base class for Pulse.

    Args:
        x (float): Rotation angle about x-axis in radians.
        y (float): Rotation angle about y-axis in radians.
        z (float): Rotation angle about z-axis in radians.
    """
    axes = 'xyz'
    indices = np.arange(3)
    indices.flags.writeable = False

    def __init__(self, x=None, y=None, z=None):

        self.__angles = np.zeros(3, dtype=np.float64)
        self.has_angle = np.zeros(3, dtype=np.float64)
        self._naxes = None
        self._flip = False

        self.x = x
        self.y = y
        self.z = z

    def set_angle(self, axis, angle):
        """
        Set rotation angle ``angle`` about axis ``axis``.

        Args:
            axis (str): Axis of the rotation.
            angle (float): Rotation angle in radians.

        Returns:

        """
        if axis not in self.axes:
            raise ValueError(f'Wrong axis format: {axis}')
        ind = _rot[axis]
        angle = _check_angle(angle)

        self._angles[ind] = angle
        self.has_angle = self.__angles.astype(bool)
        self._flip = None
        self._naxes = None

    def check_flip(self):
        """
        Check if the rotation is about single cartesian axis by an angle :math:`\pi`.
        """
        self._flip = (self.naxes == 1) and (np.isclose(self._angles[self.has_angle], np.pi))

    @property
    def naxes(self):
        """
        int: Number of axes the rotation is defined for.
        """
        if self._naxes is None:
            self._naxes = self.has_angle.sum()
        return self._naxes

    @property
    def flip(self):
        """bool: True if the angle == pi."""
        if self._flip is None:
            self.check_flip()
        return self._flip

    @property
    def _angles(self):
        return self.__angles

    @_angles.setter
    def _angles(self, angles):
        self.__angles[:] = [_check_angle(a) for a in angles]
        self.has_angle = self.__angles.astype(bool)
        self._flip = None
        self._naxes = None

    @property
    def x(self):
        """float: Angle of rotation of the spin about x axis in rad."""
        return self._angles[0]

    @x.setter
    def x(self, x):
        self.set_angle('x', x)

    @property
    def y(self):
        """float: Angle of rotation of the spin about y axis in rad."""
        return self._angles[1]

    @y.setter
    def y(self, y):
        self.set_angle('y', y)

    @property
    def z(self):
        """float: Angle of rotation of the spin about z axis in rad."""
        return self._angles[2]

    @z.setter
    def z(self, z):
        self.set_angle('z', z)

    def __repr__(self):
        inner_message = '(' + ', '.join(f'{x}: {getattr(self, x):.2f}' for x in 'xyz') + ')'
        return inner_message

    def generate_rotation(self, spinvec, spin_half=False):
        """
        Generate rotation matrix given spin vector.

        Args:
            spinvec (ndarray with shape (3, n, n)): Spin vector.
            spin_half (bool): True if spin vector is for a spin-1/2. Default is False.

        Returns:
            ndarray with shape (n, n): Rotation operator.
        """
        if spin_half and self.flip:
            return -1j * 2 * spinvec[self.indices[self.has_angle][0]]
        if self.naxes == 1:
            ind = self.indices[self.has_angle][0]
            return scipy.linalg.expm(-1j * spinvec[ind] * self._angles[ind])

        na = np.newaxis
        return scipy.linalg.expm(-1j * (spinvec[self.has_angle] * self._angles[self.has_angle][:, na, na]).sum(axis=0))


class Pulse(BasePulse, UserDict):
    """
    Class containing properties of each control pulse, applied to the system.

    The properties of the pulse, applied on the central spin(s) can be accessed as attributes, while bath spin
    pulses can be acessed as elements of the ``Pulse`` instance.

    Args:

        axis (str): Axis of rotation of the central spin. Can be 'x', 'y', or 'z'. Default is None.

        angle (float or str): Angle of rotation of central spin. Can be provided in rad, or as a string, containing
            fraction of pi: ``'pi'``, ``'pi/2'``, ``'2*pi'`` etc. Default is None.

        delay (float or ndarray): Delay before the pulse or array of delays with the same shape as time points.
            Default is None.

        which (array-like): Indexes of the central spins to be rotated by the pulse. Default is all.
            Separated indexes are supported only if qubit states are provided separately for all
            center spins.

        bath_names (str or array-like of str): Name or array of names of bath spin types, impacted by the bath pulse.
            Default is None.

        bath_axes (str or array-like of str): Axis of rotation or array of axes of the bath spins.
            Default is None. If ``bath_names`` is provided, but ``bath_axes`` and ``bath_angles`` are not,
            assumes the same axis and angle as the one of the central spin

        bath_angles (float or str or array-like): Angle of rotation or array of axes of rotations of the bath spins.

        x (float): Rotation angle of the central spin about x-axis in radians.

        y (float): Rotation angle of the central spin about y-axis in radians.

        z (float): Rotation angle of the central spin about z-axis in radians.

    Examples:

        >>> Pulse('x', 'pi')
        Pulse((x: 3.14, y: 0.00, z: 0.00))
        >>> Pulse('x', 'pi', bath_names=['13C', '14C'])
        Pulse((x: 3.14, y: 0.00, z: 0.00), {13C: (x: 3.14, y: 0.00, z: 0.00), 14C: (x: 3.14, y: 0.00, z: 0.00)})
        >>> import numpy as np
        >>> p = Pulse('x', 'pi', delay=np.linspace(0, 1, 5), bath_names=['13C', '14C'],
        >>>           bath_axes='x', bath_angles='pi/2')
        >>> print(p)
        Pulse((x: 3.14, y: 0.00, z: 0.00), {13C: (x: 1.57, y: 0.00, z: 0.00), 14C: (x: 1.57, y: 0.00, z: 0.00)},
        t = [0.   0.25 0.5  0.75 1.  ])
        >>> print(p['13C'])
        (x: 1.57, y: 0.00, z: 0.00)
    """

    def __init__(self, axis=None, angle=None, delay=None, which=None,
                 bath_names=None, bath_axes=None, bath_angles=None, **kwargs):
        super().__init__(**kwargs)
        self.data = {}

        if angle is not None:
            angle = _check_angle(angle)

        if axis is not None:
            self.set_angle(axis, angle)

        elif angle is not None:
            angle = np.asarray(angle)
            if angle.size == 3:
                self._angles = angle

        self._has_delay = False
        self.delay = delay
        """ndarray or float: Delay or array of delays before the pulse."""
        self.which = None
        """iterable: Indexes of the central spins to be rotated by the pulse."""

        if which is not None:
            self.which = np.array(which).reshape(-1)

        if bath_names is not None:
            bath_names = np.array(bath_names).reshape(-1)

            use_axes = False
            if bath_axes is not None:
                bath_axes = np.array(bath_axes).reshape(-1)
                use_axes = True

            if bath_angles is not None:
                bath_angles = np.array(bath_angles)

                if (bath_axes is None) and (bath_angles.ndim > 1 or (bath_angles.size == 3)):
                    if bath_names.size == 3 and bath_angles.size == 3:
                        warnings.warn('Ill defined bath angles format. Assuming same vector of angles for all spins.')
                    try:
                        bath_angles = bath_angles.reshape(-1, 3)
                    except ValueError:
                        raise ValueError('Wrong bath angles format')
                elif axis is not None:
                    use_axes = True
                    bath_axes = np.array(axis).reshape(-1)
                else:
                    use_axes = True
                    bath_angles = bath_angles.reshape(-1)

            else:
                bath_angles = np.array(angle).reshape(-1)

                if bath_angles.size == 3:
                    use_axes = False
                    bath_angles = bath_angles.reshape(-1, 3)
                else:
                    use_axes = True
                    bath_axes = axis
            if use_axes:
                for n, x, a in np.broadcast(bath_names, bath_axes, bath_angles):
                    self[n] = BasePulse()
                    if x is not None:
                        self[n].set_angle(x, a)
            else:
                for n, a in zip(bath_names, np.tile(bath_angles, [bath_names.size, 1])):
                    self[n] = BasePulse()
                    self[n]._angles = a

        self.bath_names = bath_names
        """ndarray: Array of names of bath spin types, impacted by the bath pulse."""

        self.bath_axes = bath_axes
        """ndarray: Array of axes of rotation of the bath spins."""

        self.bath_angles = bath_angles
        """ndarray: Array of angles of rotation of the bath spins."""

        self.rotation = None
        """ndarray: Matrix representation of the pulse for the given cluster. Generated by ``Run`` object."""

    @property
    def delay(self):
        """float or ndarray:  Delay before the pulse or array of delays with the same shape as time points."""
        return self._delay

    @delay.setter
    def delay(self, value):
        self._delay = value
        if value is not None:
            self._delay = np.asarray(value)
            self._has_delay = True
        else:
            self._has_delay = False

    def __repr__(self):

        w = f'Pulse('
        inner_message = ''

        if self.has_angle.any() is not None:
            if self.which is not None:
                inner_message += f'{self.which}: '

            x = super().__repr__()
            inner_message += x

        if self.bath_names is not None:

            bm = ''
            for k in self:
                if bm:
                    bm += ', '
                bm += f'{k}: ' + self[k].__repr__()
            if inner_message:
                inner_message += ', '
            inner_message += '{' + bm + '}'

        if self.delay is not None:
            if inner_message:
                inner_message += ', '
            inner_message += f't = {self.delay}'

        w += inner_message + ')'
        return w


_rot = {'x': 0, 'y': 1, 'z': 2}


def _check_angle(angle):
    pi = np.pi
    if angle is None:
        angle = 0
    elif isinstance(angle, str):
        angle = eval(angle)

    if np.isclose(angle, 0):
        angle = 0

    return angle


def _get_pulse(value):
    if isinstance(value, Pulse):
        value = copy.deepcopy(value)
    else:
        try:
            value = Pulse(**value)
        except TypeError:
            value = Pulse(*value)
    return value


class Sequence(UserList):
    """
    List-like object, which contains the sequence of the pulses.

    Each element is a ``Pulse`` instance, which can be generated from either the tuple with positional arguments
    or from the dictionary, or set manually.

    If delay is not provided in **all** pulses in the sequence, assume equispaced pulse sequence:

        t - pulse - 2t - pulse - 2t - ... - pulse - t

    If only **some** delays are provided, assumes 0 delay in the pulses without delay provided.


    Examples:
        >>> import numpy as np
        >>> Sequence([('x', np.pi, 0),
        >>>           {'axis': 'y', 'angle': 'pi', 'delay': np.linspace(0, 1, 3), 'bath_names': '13C'},
        >>>           Pulse('x', 'pi', 1)])
        [Pulse((x, 3.14), t = 0), Pulse((y, 3.14), {13C: (y, 3.14)}, t = [0.  0.5 1. ]), Pulse((x, 3.14), t = 1)]
    """

    def __init__(self, t=None):
        if t is not None:
            t = [_get_pulse(x) for x in t]

        super().__init__(t)

        # self.delays = None
        # """list or None: List with delays before each pulse or None if equispaced.
        # Generated by ``.generate_pulses`` method."""
        # self.rotations = None
        # """list: List with matrix representations of the rotation from each pulse.
        # Generated by ``.generate_pulses`` method."""

    def __setitem__(self, key, value):
        self.data[key] = _get_pulse(value)

    def append(self, item):
        self.data.append(_get_pulse(item))



import warnings

import numpy as np
from numba import jit
from numba.typed import List


def rotmatrix(initial_vector, final_vector):
    r"""
    Generate 3D rotation matrix which applied on initial vector will produce vector, aligned with final vector.

    Examples:

        >>> R = rotmatrix([0,0,1], [1,1,1])
        >>> R @ np.array([0,0,1])
        array([0.577, 0.577, 0.577])

    Args:
        initial_vector (ndarray with shape(3, )): Initial vector.
        final_vector (ndarray with shape (3, )): Final vector.

    Returns:
        ndarray with shape (3, 3): Rotation matrix.
    """

    iv = np.asarray(initial_vector)
    fv = np.asarray(final_vector)
    a = iv / np.linalg.norm(iv)
    b = fv / np.linalg.norm(fv)  # Final vector

    c = a @ b  # Cosine between vectors
    # if they're antiparallel
    if c == -1.:
        raise ValueError('Vectors are antiparallel')

    v = np.cross(a, b)
    screw_v = np.array([[0, -v[2], v[1]], [v[2], 0, -v[0]], [-v[1], v[0], 0]])
    r = np.eye(3) + screw_v + np.dot(screw_v, screw_v) / (1 + c)

    return r


@jit(cache=True, nopython=True)
def expand(matrix, i, dim):
    """
    Expand matrix M from it's own dimensions to the total Hilbert space.

    Args:
        matrix (ndarray with shape (dim[i], dim[i])): Inital matrix.
        i (int): Index of the spin dimensions in ``dim`` parameter.
        dim (ndarray): Array pf dimensions of all spins present in the cluster.

    Returns:
        ndarray with shape (prod(dim), prod(dim)): Expanded matrix.
    """
    dbefore = dim[:i].prod()
    dafter = dim[i + 1:].prod()

    expanded_matrix = np.kron(np.kron(np.eye(dbefore, dtype=np.complex128), matrix),
                              np.eye(dafter, dtype=np.complex128))

    return expanded_matrix


def partial_inner_product(avec, total, dimensions, index=-1):
    r"""
    Returns partial inner product :math:`\ket{b}=\bra{a}\ket{\psi}`, where :math:`\ket{a}` provided by
    ``avec`` contains degrees of freedom to be "traced out" and :math:`\ket{\psi}` provided by ``total``
    is the total statevector.

    Args:
        avec (ndarray with shape (a,)):
        total (ndarray with shape (a*b,)):
        dimensions (ndarray with shape (n,)):
        index ():

    Returns:

    """
    if len(total.shape) == 1:
        matrix = np.moveaxis(total.reshape(dimensions), index, -1)
        matrix = matrix.reshape([np.prod(np.delete(dimensions, index)), dimensions[index]])
    else:
        total = total.reshape(total.shape[0], *dimensions)
        matrix = np.moveaxis(total, index, -1)
        matrix = matrix.reshape([total.shape[0], np.prod(np.delete(dimensions, index)), dimensions[index]])
    return avec @ matrix


@jit(cache=True, nopython=True)
def shorten_dimensions(dimensions, central_number):
    """
    Combine the dimensions, corresponding to the central spins.

    Args:
        dimensions (ndarray with shape (n, )): Array of the dimensions of the spins in the cluster.
        central_number (int): Number of central spins.

    Returns:
        ndarray with shape (n - central_number): Array of the shortened dimensions;

    """
    if central_number > 1:
        shortdims = dimensions[:-central_number + 1].copy()
        # reduced dimension so all central spin dimensions are gathered in one
        shortdims[-1] = np.prod(dimensions[-central_number:])
    else:
        shortdims = dimensions
    return shortdims


@jit(cache=True, nopython=True)
def outer(s1, s2):
    """
    Outer product of two complex vectors :math:`\ket{s_1}\bra{s_2}`.

    Args:
        s1 (ndarray with shape (n, )): First vector.
        s2 (ndarray with shape (m, )): Second vector.

    Returns:
        ndarray with shape (n, m): Outer product.
    """
    return np.outer(s1, s2.conj())


@jit(cache=True, nopython=True)
def tensor_vdot(tensor, ivec):
    """
    Compute product of the tensor and spin vector.

    Args:
        tensor (ndarray with shape (3, 3)): Tensor in real space.
        ivec (ndarray with shape (3, n, n)): Spin vector.

    Returns:
        ndarray with shape (3, n, n): Right-side tensor vector product :math:`Tv`.
    """
    result = np.zeros((tensor.shape[1], *ivec.shape[1:]), dtype=ivec.dtype)
    for i, row in enumerate(tensor):
        for j, a_ij in enumerate(row):
            result[i] += a_ij * ivec[j]
    return result


@jit(cache=True, nopython=True)
def vvdot(vec_1, vec_2):
    """
    Compute product of two spin vectors.

    Args:
        vec_1 (ndarray with shape (3, N, N)): First spin vector.
        vec_2 (ndarray with shape (3, N, N)): Second spin vector.

    Returns:
        ndarray with shape (N, N): Product of two vectors.
    """
    result = np.zeros(vec_1.shape[1:], vec_1.dtype)
    for v1, v2 in zip(vec_1, vec_2):
        result += v1 @ v2
    return result


def rotate_tensor(tensor, rotation=None, style='col'):
    """
    Rootate tensor in real space, given rotation matrix.

    Args:
        tensor (ndarray with shape (3, 3)): Tensor to be rotated.
        rotation (ndarray with shape (3, 3)): Rotation matrix.
        style (str): Can be 'row' or 'col'. Determines how rotation matrix is initialized.

    Returns:
        ndarray with shape (3, 3): Rotated tensor.
    """
    if rotation is None:
        return tensor
    if style.lower == 'row':
        rotation = rotation.T
    if np.isclose(np.linalg.inv(rotation), rotation.T, rtol=1e-04).all():
        invrot = rotation.T
    else:
        warnings.warn(f"Rotation {rotation} changes distances. Is that desired behavior?", stacklevel=2)
        invrot = np.linalg.inv(rotation)
    tensor_rotation = np.matmul(tensor, rotation)
    res = np.matmul(invrot, tensor_rotation)
    #  Suppress very small deviations
    res[np.isclose(res, 0)] = 0
    return (res + np.swapaxes(res, -1, -2)) / 2


def rotate_coordinates(xyz, rotation=None, cell=None, style='col'):
    """
    Rootate coordinates in real space, given rotation matrix.

    Args:
        xyz (ndarray with shape (..., 3)): Array of coordinates.
        rotation (ndarray with shape (3, 3)): Rotation matrix.
        cell (ndarray with shape (3, 3)): Cell matrix if coordinates are given in cell coordinates.
        style (str): Can be 'row' or 'col'. Determines how rotation matrix and cell matrix are initialized.

    Returns:
        ndarray with shape (..., 3)): Array of rotated coordinates.
    """
    if style.lower() == 'row':
        if rotation is not None:
            rotation = rotation.T
        if cell is not None:
            cell = cell.T
    if cell is not None:
        xyz = np.einsum('jk,...k->...j', cell, xyz)
    if rotation is not None:
        if np.isclose(np.linalg.inv(rotation), rotation.T).all():
            invrot = rotation.T
        else:
            warnings.warn(f"Rotation {rotation} changes distances. Is that desired behavior?", stacklevel=2)
            invrot = np.linalg.inv(rotation)

        xyz = np.einsum('jk,...k->...j', invrot, xyz)
    #  Suppress very small deviations
    xyz[np.isclose(xyz, 0)] = 0

    return xyz


def normalize(vec):
    """
    Normalize vector to 1.

    Args:
        vec (ndarray with shape (n, )): Vector to be normalized.

    Returns:
        ndarray with shape (n, ): Normalized vector.
    """
    vec = np.asarray(vec, dtype=np.complex128)
    return vec / np.linalg.norm(vec)


def _add_args(after):
    def prepare_with_args(func):
        if func.__doc__ is None:
            func.__doc__ = after
        else:
            func.__doc__ = func.__doc__ + after
        return func

    return prepare_with_args


@jit(cache=True, nopython=True)
def vec_tensor_vec(v1, tensor, v2):
    """
    Compute product v @ T @ v.
    Args:
        v1 (ndarray with shape (3, n, n)): Leftmost expanded spin vector.
        tensor (ndarray with shape (3, 3)): 3x3 interaction tensor in real space.
        v2 (ndarray with shape (3, n, n)): Rightmost expanded spin vector.

    Returns:
        ndarray with shape (n, n): Product :math:`vTv`.

    """
    t_vec = tensor_vdot(tensor, v2)
    return vvdot(v1, t_vec)


@jit(cache=True, nopython=True)
def gen_state_list(states, dims):
    """
    Generate list of states from :math:`S_z` projections of the pure states.

    Args:
        states (ndarray with shape (n,)): Array of :math:`S_z` projections.
        dims (ndarray with shape (n,)): Array of the dimensions of the spins in the cluster.

    Returns:
        List: list of state vectors.

    """
    list_of_vectors = List()
    for s, d in zip(states, dims):
        list_of_vectors.append(vector_from_s(s, d))
    return list_of_vectors


@jit(cache=True, nopython=True)
def vector_from_s(s, d):
    """
    Generate vector state from :math:`S_z` projection.

    Args:
        s (float): :math:`S_z` projection.
        d (int): Dimensions of the given spin.

    Returns:
        ndarray with shape (d, ): State vector of a pure state.
    """
    vec_nucleus = np.zeros(d, dtype=np.complex128)
    state_number = np.int32((d - 1) / 2 - s)
    vec_nucleus[state_number] = 1
    return vec_nucleus



from pycce.sm import _smc
from pycce.sm import dimensions_spinvectors
from pycce.utilities import expand

from .base import Hamiltonian
from .functions import *


def bath_hamiltonian(bath, mfield):
    r"""
    Compute hamiltonian containing only the bath spins.

    Args:
        bath (BathArray):
            array of all bath spins in the cluster.

        mfield (ndarray with shape (3,) or func):
            Magnetic field of type ``mfield = np.array([Bx, By, Bz])``
            or callable with signature ``mfield(pos)``, where ``pos`` is ndarray with shape (3,) with the
            position of the spin.

    Returns:
        Hamiltonian: Hamiltonian of the given cluster without qubit.
    """
    dims, vectors = dimensions_spinvectors(bath, central_spin=None)
    clusterint = bath_interactions(bath, vectors) + custom_hamiltonian(bath, dims)

    for ivec, n in zip(vectors, bath):
        if callable(mfield):
            clusterint += expanded_single(ivec, n.gyro, mfield(n.xyz), n.Q, n.detuning)
        else:
            clusterint += expanded_single(ivec, n.gyro, mfield, n.Q, n.detuning)

    return Hamiltonian(dims, vectors, data=clusterint)


def total_hamiltonian(bath, center, mfield):
    """
    Compute total Hamiltonian of the given cluster.

    Args:
        bath (BathArray): Array of bath spins.

        center(CenterArray): Array of central spins.

        mfield (ndarray with shape (3,) or func):
            Magnetic field of type ``mfield = np.array([Bx, By, Bz])``
            or callable with signature ``mfield(pos)``, where ``pos`` is ndarray with shape (3, ) with the
            position of the spin.

    Returns:
        Hamiltonian: hamiltonian of the given cluster, including central spin.

    """

    dims, vectors = dimensions_spinvectors(bath, central_spin=center)

    totalh = bath_interactions(bath, vectors)

    ncenters = len(center)

    for i, c in enumerate(center):
        if callable(mfield):
            totalh += self_central(vectors[bath.size + i], mfield(c.xyz), c.zfs, c.gyro, c.detuning)
        else:
            # print('svec', vectors[bath.size + i].flags['C_CONTIGUOUS'])
            # print('tensor', c.zfs.flags['C_CONTIGUOUS'])
            totalh += self_central(vectors[bath.size + i], mfield, c.zfs, c.gyro, c.detuning)

    totalh += center_interactions(center, vectors[bath.size:])
    totalh += custom_hamiltonian(center, dims, offset=bath.size)
    totalh += custom_hamiltonian(bath, dims)

    for ivec, n in zip(vectors, bath):
        if callable(mfield):
            hsingle = expanded_single(ivec, n.gyro, mfield(n.xyz), n.Q, n.detuning)
        else:
            hsingle = expanded_single(ivec, n.gyro, mfield, n.Q, n.detuning)

        hhyperfine = 0

        for i in range(ncenters):
            if ncenters == 1:
                hf = n.A
            else:
                hf = n.A[i]

            hhyperfine += hyperfine(hf, vectors[bath.size + i], ivec)

        totalh += hsingle + hhyperfine

    return Hamiltonian(dims, vectors, data=totalh)


def central_hamiltonian(center, magnetic_field, hyperfine=None, bath_state=None):
    r"""
    Compute Hamiltonian, containing only central spin.

    Args:
        center (CenterArray or Center): Center spin.

        magnetic_field (ndarray with shape (3,) or func):
            Magnetic field of type ``magnetic_field = np.array([Bx, By, Bz])``
            or callable with signature ``magnetic_field(pos)``, where ``pos`` is ndarray with shape (3,) with the
            position of the spin.

        hyperfine (ndarray with shape (..., n, 3, 3)): Array of hyperfine tensors of bath spins.

        bath_state (ndarray with shape (n, )): Array of :math:`S_z` projections of bath spins.

    Returns:
        Hamiltonian: Central spin Hamiltonian.

    """
    dims, vectors = dimensions_spinvectors(central_spin=center)
    try:
        ncenters = len(center)
        single_center = False
    except TypeError:
        single_center = True
        ncenters = None

    if single_center:
        if callable(magnetic_field):
            totalh = self_central(vectors[0], magnetic_field(center.xyz),
                                  center.zfs, center.gyro, center.detuning)
        else:
            # print('svec', vectors[0].flags['C_CONTIGUOUS'])
            # print('tensor', center.zfs.flags['C_CONTIGUOUS'])
            totalh = self_central(vectors[0], magnetic_field,
                                  center.zfs, center.gyro, center.detuning)
        if hyperfine is not None and bath_state is not None:
            totalh += overhauser_central(vectors[0], hyperfine, bath_state)
        return totalh

    totalh = 0

    for i, c in enumerate(center):
        if callable(magnetic_field):
            totalh += self_central(vectors[i], magnetic_field(c.xyz), c.zfs, c.gyro)
        else:
            # print('svec', vectors[i].flags['C_CONTIGUOUS'])
            # print('tensor', c.zfs.flags['C_CONTIGUOUS'])
            totalh += self_central(vectors[i], magnetic_field, c.zfs, c.gyro)

        if hyperfine is not None and bath_state is not None:
            if ncenters == 1:
                hf = hyperfine
            else:
                hf = hyperfine[..., i, :, :]

            totalh += overhauser_central(vectors[i], hf, bath_state)

    totalh += center_interactions(center, vectors)
    totalh += custom_hamiltonian(center, dims)

    return Hamiltonian(dims, vectors, data=totalh)


def custom_hamiltonian(spins, dims=None, offset=0):
    """
    Custom addition to the Hamiltonian from the spins in the given array.

    Args:
        spins (BathArray or CenterArray): Array of the spins.
        dims (ndarray with shape (n, )): Dimensions of all spins in the cluster.
        offset (int): Index of the dimensions of the first spin from array in ``dims``. Default 0.

    Returns:
        ndarray with shape (N, N): Addition to the Hamiltonian.
    """
    if dims is None:
        dims = spins.dim
    add = 0
    for index, b in enumerate(spins):
        if b.h:
            add += custom_single(b.h, index + offset, dims)

    return add


def custom_single(h, index, dims):
    """
    Custom addition to the Hamiltonian from the dictionary with the parameters.

    Args:
        h (dict): Dictionary with coupling parameters.
        index (int): Index of the spin in ``dims``.
        dims (ndarray with shape (n, )): Dimensions of all spins in the cluster.

    Returns:
        ndarray with shape (N, N): Addition to the Hamiltonian.

    """
    sm = _smc[(dims[index] - 1) / 2]
    ham = 0
    for key in h:
        if isinstance(key, str):
            add = None
            for sym in key:
                add = h[key] * getattr(sm, sym) if add is None else np.matmul(add, getattr(sm, sym))
        else:
            add = h[key] * sm.stev(*key)

        ham += add
    return expand(ham, index, dims)
